import os

import pytorch_kinematics as pk

TEST_DIR = os.path.dirname(__file__)

def test_limits():
    chain = pk.build_serial_chain_from_urdf(open(os.path.join(TEST_DIR, "kuka_iiwa.urdf")).read(), "lbr_iiwa_link_7")
    iiwa_low_individual = []
    iiwa_high_individual = []
    for joint in chain.get_joints():
        # Default velocity and effort limits for the iiwa arm
        assert joint.velocity_limits == (-10, 10)
        assert joint.effort_limits == (-300, 300)
        iiwa_low_individual.append(joint.limits[0])
        iiwa_high_individual.append(joint.limits[1])
    iiwa_low, iiwa_high = chain.get_joint_limits()
    assert iiwa_low == iiwa_low_individual
    assert iiwa_high == iiwa_high_individual
    chain = pk.build_chain_from_urdf(open(os.path.join(TEST_DIR, "joint_limit_robot.urdf")).read())
    nums = []
    for joint in chain.get_joints():
        # Slice off the "joint" prefix to get just the number of the joint
        num = int(joint.name[5])
        nums.append(num)
        # This robot is defined specifically to test joint limits. For joint
        # number `num`, it sets lower, upper, velocity, and effort limits to
        # `num+8`, `num+9`, `num`, and `num+4` respectively
        assert joint.limits == (num + 8, num + 9)
        assert joint.velocity_limits == (-num, num)
        assert joint.effort_limits == (-(num + 4), num + 4)
    low, high = chain.get_joint_limits()
    v_low, v_high = chain.get_joint_velocity_limits()
    e_low, e_high = chain.get_joint_effort_limits()
    assert low == [x + 8 for x in nums]
    assert high == [x + 9 for x in nums]
    assert v_low == [-x for x in nums]
    assert v_high == [x for x in nums]
    assert e_low == [-(x + 4) for x in nums]
    assert e_high == [x + 4 for x in nums]


def test_empty_limits():
    chain = pk.build_chain_from_urdf(open(os.path.join(TEST_DIR, "joint_no_limit_robot.urdf")).read())
    nums = []
    for joint in chain.get_joints():
        # Slice off the "joint" prefix to get just the number of the joint
        num = int(joint.name[5])
        nums.append(num)
        # This robot is defined specifically to test joint limits. For joint
        # number `num`, it sets velocity, and effort limits to
        # `num`, and `num+4` respectively, and leaves the lower and upper
        # limits undefined
        assert joint.limits == (0, 0)
        assert joint.velocity_limits == (-num, num)
        assert joint.effort_limits == (-(num + 4), num + 4)
    low, high = chain.get_joint_limits()
    v_low, v_high = chain.get_joint_velocity_limits()
    e_low, e_high = chain.get_joint_effort_limits()
    assert low == [0] * len(nums)
    assert high == [0] * len(nums)
    assert v_low == [-x for x in nums]
    assert v_high == [x for x in nums]
    assert e_low == [-(x + 4) for x in nums]
    assert e_high == [x + 4 for x in nums]


if __name__ == "__main__":
    test_limits()
    test_empty_limits()



import os

import torch
import pytorch_kinematics as pk
import pytorch_seed

try:
    import pybullet_data
    URDF_PATH = os.path.join(pybullet_data.getDataPath(), "kuka_iiwa/model.urdf")
except ImportError:
    URDF_PATH = os.path.join(os.path.dirname(__file__), "kuka_iiwa.urdf")


def test_ik_solutions_within_joint_limits():
    pytorch_seed.seed(2)
    device = "cuda" if torch.cuda.is_available() else "cpu"
    chain = pk.build_serial_chain_from_urdf(open(URDF_PATH).read(), "lbr_iiwa_link_7")
    chain = chain.to(device=device)

    lim = torch.tensor(chain.get_joint_limits(), device=device)
    M = 100
    goal_q = torch.rand(M, lim.shape[1], device=device) * (lim[1] - lim[0]) + lim[0]
    goal_tf = chain.forward_kinematics(goal_q)

    num_retries = 10
    ik = pk.PseudoInverseIK(chain, max_iterations=50, num_retries=num_retries,
                            joint_limits=lim.T,
                            early_stopping_any_converged=True,
                            early_stopping_no_improvement="all",
                            debug=False, lr=0.2,
                            enforce_joint_limits=True)

    sol = ik.solve(goal_tf)
    print("IK solved %d / %d goals" % (sol.converged_any.sum(), M))

    # All solutions must respect joint limits
    q_all = sol.solutions
    eps = 1e-6
    low = chain.low
    high = chain.high
    assert (q_all >= low - eps).all(), \
        f"Solutions below lower limits: max violation = {(low - q_all).clamp(min=0).max()}"
    assert (q_all <= high + eps).all(), \
        f"Solutions above upper limits: max violation = {(q_all - high).clamp(min=0).max()}"

    # Convergence rate should be high since these goals are reachable
    converged_rate = sol.converged_any.sum().item() / M
    print(f"Convergence rate: {converged_rate:.1%}")
    assert converged_rate > 0.8, f"Convergence rate too low: {converged_rate:.1%}"


def test_enforce_joint_limits_flag():
    """Test that enforce_joint_limits=False disables limit enforcement."""
    pytorch_seed.seed(2)
    chain = pk.build_serial_chain_from_urdf(open(URDF_PATH).read(), "lbr_iiwa_link_7")

    lim = torch.tensor(chain.get_joint_limits())
    M = 50
    goal_q = torch.rand(M, lim.shape[1]) * (lim[1] - lim[0]) + lim[0]
    goal_tf = chain.forward_kinematics(goal_q)

    ik = pk.PseudoInverseIK(chain, max_iterations=30, num_retries=10,
                            joint_limits=lim.T,
                            early_stopping_any_converged=True,
                            early_stopping_no_improvement="all",
                            debug=False, lr=0.2,
                            enforce_joint_limits=False)

    sol = ik.solve(goal_tf)
    # With enforce_joint_limits=False, some solutions may exceed limits
    q_all = sol.solutions
    has_violation = ((q_all < chain.low).any() | (q_all > chain.high).any()).item()
    # We don't assert violations must exist (they might not), but just check it ran without error
    print(f"enforce_joint_limits=False: violations present = {has_violation}")


if __name__ == "__main__":
    test_ik_solutions_within_joint_limits()
    test_enforce_joint_limits_flag()



from functools import lru_cache
from typing import Optional, Sequence

import copy
import numpy as np
import torch

import pytorch_kinematics.transforms as tf
from pytorch_kinematics.frame import Frame, Link, Joint
from pytorch_kinematics.transforms.rotation_conversions import axis_and_angle_to_matrix_44, axis_and_d_to_pris_matrix


def get_n_joints(th):
    """

    Args:
        th: A dict, list, numpy array, or torch tensor of joints values. Possibly batched

    Returns: The number of joints in the input

    """
    if isinstance(th, torch.Tensor) or isinstance(th, np.ndarray):
        return th.shape[-1]
    elif isinstance(th, list) or isinstance(th, dict):
        return len(th)
    else:
        raise NotImplementedError(f"Unsupported type {type(th)}")


def get_batch_size(th):
    if isinstance(th, torch.Tensor) or isinstance(th, np.ndarray):
        return th.shape[0]
    elif isinstance(th, dict):
        elem_shape = get_dict_elem_shape(th)
        return elem_shape[0]
    elif isinstance(th, list):
        # Lists cannot be batched. We don't allow lists of lists.
        return 1
    else:
        raise NotImplementedError(f"Unsupported type {type(th)}")


def ensure_2d_tensor(th, dtype, device):
    if not torch.is_tensor(th):
        th = torch.tensor(th, dtype=dtype, device=device)
    if len(th.shape) <= 1:
        N = 1
        th = th.reshape(1, -1)
    else:
        N = th.shape[0]
    return th, N


def get_dict_elem_shape(th_dict):
    elem = th_dict[list(th_dict.keys())[0]]
    if isinstance(elem, np.ndarray):
        return elem.shape
    elif isinstance(elem, torch.Tensor):
        return elem.shape
    else:
        return ()


class Chain:
    """
    Robot model that may be constructed from different descriptions via their respective parsers.
    Fundamentally, a robot is modelled as a chain (not necessarily serial) of frames, with each frame
    having a physical link and a number of child frames each connected via some joint.
    """

    def __init__(self, root_frame, dtype=torch.float32, device="cpu"):
        self._root = root_frame
        self.dtype = dtype
        self.device = device

        self.identity = torch.eye(4, device=self.device, dtype=self.dtype).unsqueeze(0)

        low, high = self.get_joint_limits()
        self.low = torch.tensor(low, device=self.device, dtype=self.dtype)
        self.high = torch.tensor(high, device=self.device, dtype=self.dtype)

        # As we traverse the kinematic tree, each frame is assigned an index.
        # We use this index to build a flat representation of the tree.
        # parents_indices and joint_indices all use this indexing scheme.
        # The root frame will be index 0 and the first frame of the root frame's children will be index 1,
        # then the child of that frame will be index 2, etc. In other words, it's a depth-first ordering.
        self.parents_indices = []  # list of indices from 0 (root) to the given frame
        self.joint_indices = []
        self.n_joints = len(self.get_joint_parameter_names())
        self.axes = torch.zeros([self.n_joints, 3], dtype=self.dtype, device=self.device)
        self.link_offsets = []
        self.joint_offsets = []
        self.joint_type_indices = []
        queue = []
        queue.insert(-1, (self._root, -1, 0))  # the root has no parent so we use -1.
        idx = 0
        self.frame_to_idx = {}
        self.idx_to_frame = {}
        direct_parents = []
        frame_depths = []
        while len(queue) > 0:
            root, parent_idx, depth = queue.pop(0)
            direct_parents.append(parent_idx)
            frame_depths.append(depth)
            name_strip = root.name.strip("\n")
            self.frame_to_idx[name_strip] = idx
            self.idx_to_frame[idx] = name_strip
            if parent_idx == -1:
                self.parents_indices.append([idx])
            else:
                self.parents_indices.append(self.parents_indices[parent_idx] + [idx])

            is_fixed = root.joint.joint_type == 'fixed'

            if root.link.offset is None:
                self.link_offsets.append(None)
            else:
                self.link_offsets.append(root.link.offset.get_matrix())

            if root.joint.offset is None:
                self.joint_offsets.append(None)
            else:
                self.joint_offsets.append(root.joint.offset.get_matrix())

            if is_fixed:
                self.joint_indices.append(-1)
            else:
                jnt_idx = self.get_joint_parameter_names().index(root.joint.name)
                self.axes[jnt_idx] = root.joint.axis
                self.joint_indices.append(jnt_idx)

            # these are integers so that we can use them as indices into tensors
            # FIXME: how do we know the order of these types in C++?
            self.joint_type_indices.append(Joint.TYPES.index(root.joint.joint_type))

            for child in root.children:
                queue.append((child, idx, depth + 1))

            idx += 1
        self.joint_type_indices = torch.tensor(self.joint_type_indices)
        self.joint_indices = torch.tensor(self.joint_indices)
        # We need to use a dict because torch.compile doesn't list lists of tensors
        self.parents_indices = [torch.tensor(p, dtype=torch.long, device=self.device) for p in self.parents_indices]

        # Precomputed structures for torch.compile-compatible FK kernel
        self._num_frames = idx
        self._direct_parent_idx = torch.tensor(direct_parents, dtype=torch.long, device=self.device)

        # BFS levels: group frame indices by depth for level-by-level traversal
        max_depth = max(frame_depths) if frame_depths else 0
        self._bfs_levels = []
        for d in range(max_depth + 1):
            level_frames = [i for i, fd in enumerate(frame_depths) if fd == d]
            self._bfs_levels.append(torch.tensor(level_frames, dtype=torch.long, device=self.device))

        # Static offsets: pre-multiply link_offset @ joint_offset per frame, identity where None
        eye4 = torch.eye(4, dtype=self.dtype, device=self.device)
        static_offsets = []
        for i in range(self._num_frames):
            lo = self.link_offsets[i] if self.link_offsets[i] is not None else eye4.unsqueeze(0)
            jo = self.joint_offsets[i] if self.joint_offsets[i] is not None else eye4.unsqueeze(0)
            static_offsets.append((lo @ jo).squeeze(0))
        self._static_offsets = torch.stack(static_offsets, dim=0)  # (num_frames, 4, 4)

        # Clamped joint indices for safe tensor indexing (fixed joints: -1 → 0)
        self._joint_indices_clamped = self.joint_indices.clamp(min=0)

        # Joint type flags for skipping unnecessary computation in FK
        self._has_revolute = bool((self.joint_type_indices == 1).any())
        self._has_prismatic = bool((self.joint_type_indices == 2).any())

    def to(self, dtype=None, device=None):
        if dtype is not None:
            self.dtype = dtype
        if device is not None:
            self.device = device
        self._root = self._root.to(dtype=self.dtype, device=self.device)

        self.identity = self.identity.to(device=self.device, dtype=self.dtype)
        self.parents_indices = [p.to(dtype=torch.long, device=self.device) for p in self.parents_indices]
        self.joint_type_indices = self.joint_type_indices.to(dtype=torch.long, device=self.device)
        self.joint_indices = self.joint_indices.to(dtype=torch.long, device=self.device)
        self.axes = self.axes.to(dtype=self.dtype, device=self.device)
        self.link_offsets = [l if l is None else l.to(dtype=self.dtype, device=self.device) for l in self.link_offsets]
        self.joint_offsets = [j if j is None else j.to(dtype=self.dtype, device=self.device) for j in
                              self.joint_offsets]
        self.low = self.low.to(dtype=self.dtype, device=self.device)
        self.high = self.high.to(dtype=self.dtype, device=self.device)

        self._direct_parent_idx = self._direct_parent_idx.to(device=self.device)
        self._bfs_levels = [l.to(device=self.device) for l in self._bfs_levels]
        self._static_offsets = self._static_offsets.to(dtype=self.dtype, device=self.device)
        self._joint_indices_clamped = self._joint_indices_clamped.to(device=self.device)

        return self

    def __str__(self):
        return str(self._root)

    @staticmethod
    def _find_frame_recursive(name, frame: Frame) -> Optional[Frame]:
        for child in frame.children:
            if child.name == name:
                return child
            ret = Chain._find_frame_recursive(name, child)
            if not ret is None:
                return ret
        return None

    def find_frame(self, name) -> Optional[Frame]:
        if self._root.name == name:
            return self._root
        return self._find_frame_recursive(name, self._root)

    @staticmethod
    def _find_link_recursive(name, frame) -> Optional[Link]:
        for child in frame.children:
            if child.link.name == name:
                return child.link
            ret = Chain._find_link_recursive(name, child)
            if not ret is None:
                return ret
        return None

    @staticmethod
    def _get_joints(frame, exclude_fixed=True):
        joints = []
        if exclude_fixed and frame.joint.joint_type != "fixed":
            joints.append(frame.joint)
        for child in frame.children:
            joints.extend(Chain._get_joints(child))
        return joints

    def get_joints(self, exclude_fixed=True):
        joints = self._get_joints(self._root, exclude_fixed=exclude_fixed)
        return joints

    @lru_cache()
    def get_joint_parameter_names(self, exclude_fixed=True):
        names = []
        for j in self.get_joints(exclude_fixed=exclude_fixed):
            if exclude_fixed and j.joint_type == 'fixed':
                continue
            names.append(j.name)
        return names

    @staticmethod
    def _find_joint_recursive(name, frame):
        for child in frame.children:
            if child.joint.name == name:
                return child.joint
            ret = Chain._find_joint_recursive(name, child)
            if not ret is None:
                return ret
        return None

    def find_link(self, name) -> Optional[Link]:
        if self._root.link.name == name:
            return self._root.link
        return self._find_link_recursive(name, self._root)

    def find_joint(self, name):
        if self._root.joint.name == name:
            return self._root.joint
        return self._find_joint_recursive(name, self._root)

    @staticmethod
    def _get_joint_parent_frame_names(frame, exclude_fixed=True):
        joint_names = []
        if not (exclude_fixed and frame.joint.joint_type == "fixed"):
            joint_names.append(frame.name)
        for child in frame.children:
            joint_names.extend(Chain._get_joint_parent_frame_names(child, exclude_fixed))
        return joint_names

    def get_joint_parent_frame_names(self, exclude_fixed=True):
        names = self._get_joint_parent_frame_names(self._root, exclude_fixed)
        return sorted(set(names), key=names.index)

    @staticmethod
    def _get_frame_names(frame: Frame, exclude_fixed=True) -> Sequence[str]:
        names = []
        if not (exclude_fixed and frame.joint.joint_type == "fixed"):
            names.append(frame.name)
        for child in frame.children:
            names.extend(Chain._get_frame_names(child, exclude_fixed))
        return names

    def get_frame_names(self, exclude_fixed=True):
        names = self._get_frame_names(self._root, exclude_fixed)
        return sorted(set(names), key=names.index)

    @staticmethod
    def _get_links(frame):
        links = [frame.link]
        for child in frame.children:
            links.extend(Chain._get_links(child))
        return links

    def get_links(self):
        links = self._get_links(self._root)
        return links

    @staticmethod
    def _get_link_names(frame):
        link_names = [frame.link.name]
        for child in frame.children:
            link_names.extend(Chain._get_link_names(child))
        return link_names

    def get_link_names(self):
        names = self._get_link_names(self._root)
        return sorted(set(names), key=names.index)

    @lru_cache
    def get_frame_indices(self, *frame_names):
        return torch.tensor([self.frame_to_idx[n] for n in frame_names], dtype=torch.long, device=self.device)

    def print_tree(self, do_print=True):
        tree = str(self._root)
        if do_print:
            print(tree)
        return tree

    def forward_kinematics_tensor(self, th):
        """
        Compilable FK kernel. Computes transforms for all frames using level-by-level BFS traversal.
        Compatible with torch.compile(fullgraph=True).

        Args:
            th: (B, n_joints) joint angle tensor

        Returns: (num_frames, B, 4, 4) tensor of all frame transforms
        """
        B = th.shape[0]

        # Compute joint transforms for all joints at once, skipping unused types.
        # _has_revolute/_has_prismatic are Python bools so torch.compile specializes on them.
        jidx = self._joint_indices_clamped
        eye4 = torch.eye(4, device=th.device, dtype=th.dtype)

        if self.n_joints > 0:
            axes_expanded = self.axes.unsqueeze(0).expand(B, -1, -1)

            if self._has_revolute:
                rev_transforms = axis_and_angle_to_matrix_44(axes_expanded, th)
                rev_per_frame = rev_transforms[:, jidx].permute(1, 0, 2, 3)

            if self._has_prismatic:
                pris_transforms = axis_and_d_to_pris_matrix(axes_expanded, th)
                pris_per_frame = pris_transforms[:, jidx].permute(1, 0, 2, 3)

        if self._has_revolute and not self._has_prismatic:
            # Revolute-only: skip prismatic computation and torch.where selection
            is_rev = (self.joint_type_indices == 1).reshape(-1, 1, 1, 1)
            joint_transforms = eye4.reshape(1, 1, 4, 4).expand(self._num_frames, B, 4, 4)
            joint_transforms = torch.where(is_rev, rev_per_frame, joint_transforms)
        elif self._has_prismatic and not self._has_revolute:
            # Prismatic-only: skip revolute computation and torch.where selection
            is_pris = (self.joint_type_indices == 2).reshape(-1, 1, 1, 1)
            joint_transforms = eye4.reshape(1, 1, 4, 4).expand(self._num_frames, B, 4, 4)
            joint_transforms = torch.where(is_pris, pris_per_frame, joint_transforms)
        elif self._has_revolute and self._has_prismatic:
            # Mixed: need both
            is_rev = (self.joint_type_indices == 1).reshape(-1, 1, 1, 1)
            is_pris = (self.joint_type_indices == 2).reshape(-1, 1, 1, 1)
            joint_transforms = eye4.reshape(1, 1, 4, 4).expand(self._num_frames, B, 4, 4)
            joint_transforms = torch.where(is_rev, rev_per_frame, joint_transforms)
            joint_transforms = torch.where(is_pris, pris_per_frame, joint_transforms)
        else:
            # All fixed joints
            joint_transforms = eye4.reshape(1, 1, 4, 4).expand(self._num_frames, B, 4, 4)

        # Local transforms: static_offsets @ joint_transforms
        local_transforms = self._static_offsets.unsqueeze(1) @ joint_transforms

        # Level-by-level transform accumulation
        buffer = torch.empty(self._num_frames, B, 4, 4, device=th.device, dtype=th.dtype)

        # Level 0: root frames (no parent to compose with)
        level_indices = self._bfs_levels[0]
        buffer[level_indices] = local_transforms[level_indices]

        # Level 1+: compose with parent transforms
        for d in range(1, len(self._bfs_levels)):
            level_indices = self._bfs_levels[d]
            parents = self._direct_parent_idx[level_indices]
            buffer[level_indices] = buffer[parents] @ local_transforms[level_indices]

        return buffer

    def forward_kinematics(self, th, frame_indices: Optional = None):
        """
        Compute forward kinematics for the given joint values.

        Args:
            th: A dict, list, numpy array, or torch tensor of joints values. Possibly batched.
            frame_indices: A list of frame indices to compute transforms for. If None, all frames are computed.
                Use `get_frame_indices` to convert from frame names to frame indices.

        Returns:
            A dict of Transform3d objects for each frame.

        """
        if frame_indices is None:
            frame_indices = self.get_all_frame_indices()

        th = self.ensure_tensor(th)
        th = torch.atleast_2d(th)

        all_transforms = self.forward_kinematics_tensor(th)

        return {self.idx_to_frame[fi.item()]: tf.Transform3d(matrix=all_transforms[fi])
                for fi in frame_indices}

    def ensure_tensor(self, th):
        """
        Converts a number of possible types into a tensor. The order of the tensor is determined by the order
        of self.get_joint_parameter_names(). th must contain all joints in the entire chain.
        """
        if isinstance(th, np.ndarray):
            th = torch.tensor(th, device=self.device, dtype=self.dtype)
        elif isinstance(th, list):
            th = torch.tensor(th, device=self.device, dtype=self.dtype)
        elif isinstance(th, dict):
            # convert dict to a flat, complete, tensor of all joints values. Missing joints are filled with zeros.
            th_dict = th
            elem_shape = get_dict_elem_shape(th_dict)
            th = torch.ones([*elem_shape, self.n_joints], device=self.device, dtype=self.dtype) * torch.nan
            joint_names = self.get_joint_parameter_names()
            for joint_name, joint_position in th_dict.items():
                jnt_idx = joint_names.index(joint_name)
                th[..., jnt_idx] = joint_position
            if torch.any(torch.isnan(th)):
                msg = "Missing values for the following joints:\n"
                for joint_name, th_i in zip(self.get_joint_parameter_names(), th):
                    msg += joint_name + "\n"
                raise ValueError(msg)
        return th

    def get_all_frame_indices(self):
        frame_indices = self.get_frame_indices(*self.get_frame_names(exclude_fixed=False))
        return frame_indices

    def clamp(self, th):
        """

        Args:
            th: Joint configuration

        Returns: Always a tensor in the order of self.get_joint_parameter_names(), possibly batched.

        """
        th = self.ensure_tensor(th)
        return torch.clamp(th, self.low, self.high)

    def get_joint_limits(self):
        return self._get_joint_limits("limits")

    def get_joint_velocity_limits(self):
        return self._get_joint_limits("velocity_limits")

    def get_joint_effort_limits(self):
        return self._get_joint_limits("effort_limits")

    def _get_joint_limits(self, param_name):
        low = []
        high = []
        for joint in self.get_joints():
            val = getattr(joint, param_name)
            if val is None:
                # NOTE: This changes the previous default behavior of returning
                # +/- np.pi for joint limits to be more natural for both
                # revolute and prismatic joints
                low.append(-np.inf)
                high.append(np.inf)
            else:
                low.append(val[0])
                high.append(val[1])
        return low, high

    @staticmethod
    def _get_joints_and_child_links(frame):
        joint = frame.joint

        me_and_my_children = [frame.link]
        for child in frame.children:
            recursive_child_links = yield from Chain._get_joints_and_child_links(child)
            me_and_my_children.extend(recursive_child_links)

        if joint is not None and joint.joint_type != 'fixed':
            yield joint, me_and_my_children

        return me_and_my_children

    def get_joints_and_child_links(self):
        yield from Chain._get_joints_and_child_links(self._root)


class SerialChain(Chain):
    """
    A serial Chain specialization with no branches and clearly defined end effector.
    Serial chains can be generated from subsets of a Chain.
    """

    def __init__(self, chain, end_frame_name, root_frame_name="", **kwargs):
        root_frame = chain._root if root_frame_name == "" else chain.find_frame(root_frame_name)
        if root_frame is None:
            raise ValueError("Invalid root frame name %s." % root_frame_name)
        chain = Chain(root_frame, **kwargs)

        # make a copy of those frames that includes only the chain up to the end effector
        end_frame_idx = chain.get_frame_indices(end_frame_name)
        ancestors = chain.parents_indices[end_frame_idx]

        frames = []
        # first pass create copies of the ancestor nodes
        for idx in ancestors:
            this_frame_name = chain.idx_to_frame[idx.item()]
            this_frame = copy.deepcopy(chain.find_frame(this_frame_name))
            if idx == end_frame_idx:
                this_frame.children = []
            frames.append(this_frame)
        # second pass assign correct children (only the next one in the frame list)
        for i in range(len(ancestors) - 1):
            frames[i].children = [frames[i + 1]]

        # The root frame's joint and link offset describe the transform from its parent
        # to itself, which is outside this serial chain. Reset them so the root acts as
        # the origin of the new coordinate system.
        frames[0].joint = Joint()
        frames[0].link.offset = None

        self._serial_frames = frames
        super().__init__(frames[0], **kwargs)

        # Precompute data for compilable Jacobian (jacobian_tensor)
        dof_frame_indices = []
        dof_axes = []
        dof_types = []  # 1=revolute, 2=prismatic
        for f in self._serial_frames:
            if f.joint.joint_type != 'fixed':
                fi = self.frame_to_idx[f.name]
                dof_frame_indices.append(fi)
                dof_axes.append(f.joint.axis)
                jtype = 1 if f.joint.joint_type == 'revolute' else 2
                dof_types.append(jtype)

        self._serial_dof_frame_indices = torch.tensor(dof_frame_indices, dtype=torch.long, device=self.device)
        if dof_axes:
            self._serial_dof_axes = torch.stack(dof_axes, dim=0).to(device=self.device, dtype=self.dtype)  # (ndof, 3)
        else:
            self._serial_dof_axes = torch.zeros(0, 3, device=self.device, dtype=self.dtype)
        self._serial_dof_types = torch.tensor(dof_types, dtype=torch.long, device=self.device)  # (ndof,)
        self._serial_eef_frame_idx = self.frame_to_idx[self._serial_frames[-1].name]

    def to(self, dtype=None, device=None):
        super().to(dtype=dtype, device=device)
        self._serial_dof_frame_indices = self._serial_dof_frame_indices.to(device=self.device)
        self._serial_dof_axes = self._serial_dof_axes.to(dtype=self.dtype, device=self.device)
        self._serial_dof_types = self._serial_dof_types.to(device=self.device)
        return self

    def jacobian_tensor(self, th, ret_eef_pose=False, all_transforms=None):
        """
        Compilable Jacobian kernel. Computes the geometric Jacobian in the base frame.
        Compatible with torch.compile(fullgraph=True).

        Args:
            th: (B, n_joints) joint angle tensor
            ret_eef_pose: if True, also return the (B, 4, 4) end-effector pose matrix.
            all_transforms: optional pre-computed (num_frames, B, 4, 4) FK transforms.
                If provided, skips the internal FK computation. Useful to avoid redundant FK calls.

        Returns: (B, 6, ndof) geometric Jacobian, and optionally (B, 4, 4) EEF pose
        """
        if all_transforms is None:
            all_transforms = self.forward_kinematics_tensor(th)  # (num_frames, B, 4, 4)

        p_ee = all_transforms[self._serial_eef_frame_idx, :, :3, 3]  # (B, 3)

        # DOF frame transforms
        T_dof = all_transforms[self._serial_dof_frame_indices]  # (ndof, B, 4, 4)
        R_dof = T_dof[:, :, :3, :3]  # (ndof, B, 3, 3)
        p_dof = T_dof[:, :, :3, 3]   # (ndof, B, 3)

        # Joint axes in base frame: z_i = R_i @ axis_i
        axes = self._serial_dof_axes  # (ndof, 3)
        z = torch.einsum('nbij,nj->nbi', R_dof, axes)  # (ndof, B, 3)

        # Position difference: p_ee - p_i
        dp = p_ee.unsqueeze(0) - p_dof  # (ndof, B, 3)

        # Revolute: J_v = z x dp, J_w = z
        # Prismatic: J_v = z, J_w = 0
        cross = torch.cross(z, dp, dim=2)  # (ndof, B, 3)

        is_rev = (self._serial_dof_types == 1).reshape(-1, 1, 1)  # (ndof, 1, 1)

        J_v = torch.where(is_rev, cross, z)  # (ndof, B, 3)
        J_w = torch.where(is_rev, z, torch.zeros_like(z))  # (ndof, B, 3)

        # Stack to (B, 6, ndof)
        J = torch.cat([J_v, J_w], dim=2)  # (ndof, B, 6)
        J = J.permute(1, 2, 0)  # (B, 6, ndof)

        if ret_eef_pose:
            return J, all_transforms[self._serial_eef_frame_idx]
        return J

    def jacobian(self, th, locations=None, ret_eef_pose=False):
        """
        Compute the geometric Jacobian in the base frame.

        Args:
            th: Joint angles as a dict, list, numpy array, or torch tensor. Possibly batched.
            locations: (B, 3) or (3,) tool offset position relative to the end effector.
            ret_eef_pose: if True, also return the (B, 4, 4) end-effector pose matrix.

        Returns: (B, 6, ndof) Jacobian, and optionally (B, 4, 4) EEF pose
        """
        if not torch.is_tensor(th):
            th = torch.tensor(th, dtype=self.dtype, device=self.device)
        if len(th.shape) <= 1:
            th = th.reshape(1, -1)

        if locations is None and not ret_eef_pose:
            return self.jacobian_tensor(th)

        if locations is not None:
            # Compute FK once and pass to jacobian_tensor to avoid redundant FK
            all_transforms = self.forward_kinematics_tensor(th)
            J = self.jacobian_tensor(th, all_transforms=all_transforms)
            T_ee = all_transforms[self._serial_eef_frame_idx]  # (B, 4, 4)

            if isinstance(locations, tf.Transform3d):
                tool = locations
            else:
                tool = tf.Transform3d(pos=locations)
            if tool.dtype != self.dtype or tool.device != self.device:
                tool = tool.to(device=self.device, copy=True, dtype=self.dtype)
            tool_matrix = tool.get_matrix()
            T_ee_tool = T_ee @ tool_matrix

            # Tool offset only changes p_ee, affecting linear velocity of revolute joints:
            #   J_v_new[i] = z_i x (p_ee_tool - p_i) = J_v_old[i] + z_i x (p_ee_tool - p_ee)
            # Prismatic joints are unaffected (J_v = z, independent of p_ee).
            delta_p = T_ee_tool[:, :3, 3] - T_ee[:, :3, 3]  # (B, 3)

            T_dof = all_transforms[self._serial_dof_frame_indices]
            R_dof = T_dof[:, :, :3, :3]
            z = torch.einsum('nbij,nj->nbi', R_dof, self._serial_dof_axes)  # (ndof, B, 3)

            correction = torch.cross(z, delta_p.unsqueeze(0).expand_as(z), dim=2)  # (ndof, B, 3)
            is_rev = (self._serial_dof_types == 1).reshape(-1, 1, 1)
            correction = torch.where(is_rev, correction, torch.zeros_like(correction))
            J = J.clone()
            J[:, :3, :] = J[:, :3, :] + correction.permute(1, 2, 0)

            T_ee = T_ee_tool
            if ret_eef_pose:
                return J, T_ee
            return J

        # ret_eef_pose=True, no locations: get both from single FK pass
        J, T_ee = self.jacobian_tensor(th, ret_eef_pose=True)
        return J, T_ee

    def forward_kinematics(self, th, end_only: bool = True):
        """ Like the base class, except `th` only needs to contain the joints in the SerialChain, not all joints. """
        frame_indices, th = self.convert_serial_inputs_to_chain_inputs(th, end_only)

        mat = super().forward_kinematics(th, frame_indices)

        if end_only:
            return mat[self._serial_frames[-1].name]
        else:
            return mat

    def convert_serial_inputs_to_chain_inputs(self, th, end_only: bool):
        # th = self.ensure_tensor(th)
        th_b = get_batch_size(th)
        th_n_joints = get_n_joints(th)
        if isinstance(th, list):
            th = torch.tensor(th, device=self.device, dtype=self.dtype)

        if end_only:
            frame_indices = self.get_frame_indices(self._serial_frames[-1].name)
        else:
            # pass through default behavior for frame indices being None, which is currently
            # to return all frames.
            frame_indices = None
        if th_n_joints < self.n_joints:
            # if th is only a partial list of joints, assume it's a list of joints for only the serial chain.
            partial_th = th
            nonfixed_serial_frames = list(filter(lambda f: f.joint.joint_type != 'fixed', self._serial_frames))
            if th_n_joints != len(nonfixed_serial_frames):
                raise ValueError(f'Expected {len(nonfixed_serial_frames)} joint values, got {th_n_joints}.')
            th = torch.zeros([th_b, self.n_joints], device=self.device, dtype=self.dtype)
            for i, frame in enumerate(nonfixed_serial_frames):
                joint_name = frame.joint.name
                if isinstance(partial_th, dict):
                    partial_th_i = partial_th[joint_name]
                else:
                    partial_th_i = partial_th[..., i]
                k = self.frame_to_idx[frame.name]
                jnt_idx = self.joint_indices[k]
                if frame.joint.joint_type != 'fixed':
                    th[..., jnt_idx] = partial_th_i
        return frame_indices, th



import os
from timeit import default_timer as timer

import torch

import pytorch_kinematics as pk

TEST_DIR = os.path.dirname(__file__)


def test_extract_serial_chain_from_tree():
    urdf = "widowx/wx250s.urdf"
    full_urdf = os.path.join(TEST_DIR, urdf)
    chain = pk.build_chain_from_urdf(open(full_urdf, mode="rb").read())
    # full frames
    full_frame_expected = """
base_link
└── shoulder_link
    └── upper_arm_link
        └── upper_forearm_link
            └── lower_forearm_link
                └── wrist_link
                    └── gripper_link
                        └── ee_arm_link
                            ├── gripper_prop_link
                            └── gripper_bar_link
                                └── fingers_link
                                    ├── left_finger_link
                                    ├── right_finger_link
                                    └── ee_gripper_link
    """
    full_frame = chain.print_tree()
    assert full_frame_expected.strip() == full_frame.strip()

    serial_chain = pk.SerialChain(chain, "ee_gripper_link", "base_link")
    serial_frame_expected = """
base_link
└── shoulder_link
    └── upper_arm_link
        └── upper_forearm_link
            └── lower_forearm_link
                └── wrist_link
                    └── gripper_link
                        └── ee_arm_link
                            └── gripper_bar_link
                                └── fingers_link
                                    └── ee_gripper_link
    """
    serial_frame = serial_chain.print_tree()
    assert serial_frame_expected.strip() == serial_frame.strip()

    # full chain should have DOF = 8, however since we are creating just a serial chain to ee_gripper_link, should be 6
    assert chain.n_joints == 8
    assert serial_chain.n_joints == 6

    serial_chain = pk.SerialChain(chain, "gripper_prop_link", "base_link")
    serial_frame_expected = """
base_link
└── shoulder_link
    └── upper_arm_link
        └── upper_forearm_link
            └── lower_forearm_link
                └── wrist_link
                    └── gripper_link
                        └── ee_arm_link
                            └── gripper_prop_link
    """
    serial_frame = serial_chain.print_tree()
    assert serial_frame_expected.strip() == serial_frame.strip()

    serial_chain = pk.SerialChain(chain, "ee_gripper_link", "gripper_link")
    serial_frame_expected = """
gripper_link
└── ee_arm_link
    └── gripper_bar_link
        └── fingers_link
            └── ee_gripper_link
        """
    serial_frame = serial_chain.print_tree()
    assert serial_frame_expected.strip() == serial_frame.strip()
    # gripper_link's own joint connects wrist_link -> gripper_link, which is outside
    # this sub-chain, so it should not be counted. All remaining joints are fixed.
    assert serial_chain.n_joints == 0


def test_serial_chain_non_root_start():
    """Test that SerialChain with a non-root root_frame_name excludes the root's own joint (issue #62)."""
    data = ('<robot name="test_robot">'
            '<link name="link1" />'
            '<link name="link2" />'
            '<link name="link3" />'
            '<link name="link4" />'
            '<joint name="joint1" type="revolute">'
            '<origin xyz="1.0 0.0 0.0"/>'
            '<parent link="link1"/>'
            '<child link="link2"/>'
            '</joint>'
            '<joint name="joint2" type="revolute">'
            '<origin xyz="1.0 0.0 0.0"/>'
            '<parent link="link2"/>'
            '<child link="link3"/>'
            '</joint>'
            '<joint name="joint3" type="revolute">'
            '<origin xyz="1.0 0.0 0.0"/>'
            '<parent link="link3"/>'
            '<child link="link4"/>'
            '</joint>'
            '</robot>')
    full_chain = pk.build_chain_from_urdf(data)

    # SerialChain from link2 to link4 should only include joint2 and joint3, not joint1
    serial = pk.SerialChain(full_chain, "link4", "link2")
    assert "joint1" not in serial.get_joint_parameter_names()
    assert serial.get_joint_parameter_names() == ["joint2", "joint3"]

    # FK from a frame to itself should be identity
    identity_chain = pk.SerialChain(full_chain, "link2", "link2")
    tg = identity_chain.forward_kinematics(torch.zeros(0))
    assert torch.allclose(tg.get_matrix(), torch.eye(4).unsqueeze(0), atol=1e-7)

    # FK from link2 to link4 with zeros should give translation of (2, 0, 0) — two joints each offset by (1, 0, 0)
    tg = serial.forward_kinematics(torch.zeros(2))
    pos = tg.get_matrix()[0, :3, 3]
    assert torch.allclose(pos, torch.tensor([2.0, 0.0, 0.0]), atol=1e-7)

    # Compare: full chain FK at zeros, link4 relative to link2
    full_ret = full_chain.forward_kinematics(torch.zeros(3))
    full_link2 = full_ret['link2'].get_matrix()
    full_link4 = full_ret['link4'].get_matrix()
    relative = torch.linalg.inv(full_link2) @ full_link4
    assert torch.allclose(relative, tg.get_matrix(), atol=1e-7)


if __name__ == "__main__":
    test_extract_serial_chain_from_tree()
    test_serial_chain_non_root_start()



from pytorch_kinematics.chain import SerialChain
from pytorch_kinematics.transforms import Transform3d
from pytorch_kinematics.transforms import rotation_conversions
from typing import Union, Optional, Callable
import math
import torch
from matplotlib import pyplot as plt, cm as cm

# Check if torch.compile is available (PyTorch 2.0+)
_TORCH_COMPILE_AVAILABLE = hasattr(torch, 'compile') and torch.__version__ >= '2.0'


def _ik_step_kernel(m_flat: torch.Tensor, target_pos: torch.Tensor, target_wxyz: torch.Tensor,
                    J: torch.Tensor, reg_matrix: torch.Tensor, num_retries: int,
                    lm_damping: float = 0.0, task_weight: Optional[torch.Tensor] = None):
    """
    Fused IK step: delta_pose + damped least squares. Compatible with torch.compile(fullgraph=True).

    Args:
        m_flat: End-effector poses (N*M, 4, 4) where N=num_problems, M=num_retries
        target_pos: Target positions (N, 3)
        target_wxyz: Target orientations as wxyz quaternions (N, 4)
        J: Jacobian matrix (N*M, 6, DOF)
        reg_matrix: Regularization matrix (6, 6)
        num_retries: Number of retries per problem (M)
        lm_damping: Levenberg-Marquardt damping factor. When > 0, adds error-proportional
            regularization: mu = lm_damping * ||dx||^2. Large errors get more damping (stable),
            small errors get less (fast convergence). Inspired by mink's task-level LM damping.

    Returns:
        dq: Joint velocity (N*M, DOF, 1)
        dx: Pose error (N*M, 6, 1)
    """
    NM = m_flat.shape[0]
    M = num_retries
    N = NM // M
    m = m_flat.view(N, M, 4, 4)

    # delta_pose
    pos_diff = (target_pos.unsqueeze(1) - m[:, :, :3, 3]).view(-1, 3, 1)
    cur_wxyz = rotation_conversions.matrix_to_quaternion(m[:, :, :3, :3])
    diff_wxyz = rotation_conversions.quaternion_multiply(
        target_wxyz.unsqueeze(1),
        rotation_conversions.quaternion_invert(cur_wxyz))
    diff_axis_angle = rotation_conversions.quaternion_to_axis_angle(diff_wxyz)
    rot_diff = diff_axis_angle.view(-1, 3, 1)
    dx = torch.cat((pos_diff, rot_diff), dim=1)

    # Apply per-coordinate weighting: W @ J, W @ dx
    if task_weight is not None:
        w = task_weight.reshape(1, 6, 1)  # (1, 6, 1)
        J_w = w * J       # (NM, 6, DOF) weighted Jacobian
        dx_w = w * dx     # (NM, 6, 1) weighted error
    else:
        J_w = J
        dx_w = dx

    # DLS with adaptive Levenberg-Marquardt damping:
    # reg = lambda^2 * I + lm_damping * ||dx_w||^2 * I
    if lm_damping > 0.0:
        mu = lm_damping * (dx_w * dx_w).sum(dim=1, keepdim=True)  # (NM, 1, 1)
        tmpA = J_w @ J_w.transpose(1, 2) + reg_matrix + mu * torch.eye(6, device=J.device, dtype=J.dtype)
    else:
        tmpA = J_w @ J_w.transpose(1, 2) + reg_matrix
    A = torch.linalg.solve(tmpA, dx_w)
    dq = J_w.transpose(1, 2) @ A
    return dq, dx


def _ik_step_kernel_svd(m_flat: torch.Tensor, target_pos: torch.Tensor, target_wxyz: torch.Tensor,
                        J: torch.Tensor, reg_matrix: torch.Tensor, num_retries: int,
                        lm_damping: float = 0.0, task_weight: Optional[torch.Tensor] = None):
    """
    IK step using SVD-based damped least squares. Generally slower than the Cholesky-based
    kernel, but exposes singular values for selective damping if needed.

    The DLS pseudoinverse is: J^T (JJ^T + lambda^2 I)^{-1}
    Via SVD (J = U D V^T): sum_i (d_i / (d_i^2 + lambda^2)) v_i u_i^T

    Args:
        m_flat: End-effector poses (N*M, 4, 4)
        target_pos: Target positions (N, 3)
        target_wxyz: Target orientations as wxyz quaternions (N, 4)
        J: Jacobian matrix (N*M, 6, DOF)
        reg_matrix: Regularization matrix (6, 6) — only the diagonal (scalar) is used
        num_retries: Number of retries per problem (M)

    Returns:
        dq: Joint velocity (N*M, DOF, 1)
        dx: Pose error (N*M, 6, 1)
    """
    NM = m_flat.shape[0]
    M = num_retries
    N = NM // M
    m = m_flat.view(N, M, 4, 4)

    # delta_pose (same as _ik_step_kernel)
    pos_diff = (target_pos.unsqueeze(1) - m[:, :, :3, 3]).view(-1, 3, 1)
    cur_wxyz = rotation_conversions.matrix_to_quaternion(m[:, :, :3, :3])
    diff_wxyz = rotation_conversions.quaternion_multiply(
        target_wxyz.unsqueeze(1),
        rotation_conversions.quaternion_invert(cur_wxyz))
    diff_axis_angle = rotation_conversions.quaternion_to_axis_angle(diff_wxyz)
    rot_diff = diff_axis_angle.view(-1, 3, 1)
    dx = torch.cat((pos_diff, rot_diff), dim=1)

    # Apply per-coordinate weighting
    if task_weight is not None:
        w = task_weight.reshape(1, 6, 1)
        J_w = w * J
        dx_w = w * dx
    else:
        J_w = J
        dx_w = dx

    # SVD-based DLS
    regularization = reg_matrix[0, 0]  # scalar lambda^2
    if lm_damping > 0.0:
        mu = lm_damping * (dx_w * dx_w).sum(dim=1, keepdim=True).squeeze(2)  # (NM, 1)
        regularization = regularization + mu
    U, D, Vh = torch.linalg.svd(J_w)
    m_sv = D.shape[1]
    denom = D ** 2 + regularization
    prod = D / denom
    inverted = torch.diag_embed(prod)
    Vh = Vh[:, :m_sv, :]
    total = Vh.transpose(1, 2) @ inverted @ U.transpose(1, 2)
    dq = total @ dx_w
    return dq, dx


class IKSolution:
    def __init__(self, dof, num_problems, num_retries, pos_tolerance, rot_tolerance, device="cpu", dtype=None):
        self.iterations = 0
        self.device = device
        self.num_problems = num_problems
        self.num_retries = num_retries
        self.dof = dof
        self.pos_tolerance = pos_tolerance
        self.rot_tolerance = rot_tolerance

        M = num_problems
        # N x DOF tensor of joint angles; if converged[i] is False, then solutions[i] is undefined
        self.solutions = torch.zeros((M, self.num_retries, self.dof), device=self.device, dtype=dtype)
        self.remaining = torch.ones(M, dtype=torch.bool, device=self.device)

        # M is the total number of problems
        # N is the total number of attempts
        # M x N tensor of position and rotation errors
        self.err_pos = torch.zeros((M, self.num_retries), device=self.device, dtype=dtype)
        self.err_rot = torch.zeros_like(self.err_pos)
        # M x N boolean values indicating whether the solution converged (a solution could be found)
        self.converged_pos = torch.zeros((M, self.num_retries), dtype=torch.bool, device=self.device)
        self.converged_rot = torch.zeros_like(self.converged_pos)
        self.converged = torch.zeros_like(self.converged_pos)

        # M whether any position and rotation converged for that problem
        self.converged_pos_any = torch.zeros_like(self.remaining)
        self.converged_rot_any = torch.zeros_like(self.remaining)
        self.converged_any = torch.zeros_like(self.remaining)

    def update_remaining_with_keep_mask(self, keep: torch.tensor):
        self.remaining = self.remaining & keep
        return self.remaining

    def update(self, q: torch.tensor, err: torch.tensor, use_keep_mask=True, keep_mask=None):
        err = err.reshape(-1, self.num_retries, 6)
        err_pos = err[..., :3].norm(dim=-1)
        err_rot = err[..., 3:].norm(dim=-1)
        converged_pos = err_pos < self.pos_tolerance
        converged_rot = err_rot < self.rot_tolerance
        converged = converged_pos & converged_rot
        converged_any = converged.any(dim=1)

        if keep_mask is None:
            keep_mask = ~converged_any

        # stop considering problems where any converged
        qq = q.reshape(-1, self.num_retries, self.dof)

        if use_keep_mask:
            # those that have converged are no longer remaining
            self.update_remaining_with_keep_mask(keep_mask)

        # sticky convergence: only overwrite retries that haven't already converged
        # this prevents overwriting a good solution if the solver overshoots on the next step
        # use torch.where instead of boolean indexing for compile compatibility
        not_converged = ~self.converged
        not_converged_dof = not_converged.unsqueeze(-1)  # (M, num_retries, 1) for DOF broadcast
        self.solutions = torch.where(not_converged_dof, qq, self.solutions)
        self.err_pos = torch.where(not_converged, err_pos, self.err_pos)
        self.err_rot = torch.where(not_converged, err_rot, self.err_rot)
        self.converged_pos = self.converged_pos | converged_pos
        self.converged_rot = self.converged_rot | converged_rot
        self.converged = self.converged | converged
        self.converged_any = self.converged_any | converged_any

        return converged_any


# helper config sampling method
def gaussian_around_config(config: torch.Tensor, std: float) -> Callable[[int], torch.Tensor]:
    def config_sampling_method(num_configs):
        return torch.randn(num_configs, config.shape[0], dtype=config.dtype, device=config.device) * std + config

    return config_sampling_method


class LineSearch:
    def do_line_search(self, chain, q, dq, target_pos, target_wxyz, initial_dx, problem_remaining=None,
                       fk_fn=None, eef_frame_idx=None):
        raise NotImplementedError()


class BacktrackingLineSearch(LineSearch):
    def __init__(self, max_lr=1.0, decrease_factor=0.5, max_iterations=5, sufficient_decrease=0.01):
        self.initial_lr = max_lr
        self.decrease_factor = decrease_factor
        self.max_iterations = max_iterations
        self.sufficient_decrease = sufficient_decrease

    def do_line_search(self, chain, q, dq, target_pos, target_wxyz, initial_dx, problem_remaining=None,
                       fk_fn=None, eef_frame_idx=None):
        N = target_pos.shape[0]
        NM = q.shape[0]
        M = NM // N
        lr = torch.ones(NM, device=q.device) * self.initial_lr
        err = initial_dx.squeeze().norm(dim=-1)
        if problem_remaining is None:
            problem_remaining = torch.ones(N, dtype=torch.bool, device=q.device)
        remaining = torch.ones((N, M), dtype=torch.bool, device=q.device)
        # don't care about the ones that are no longer remaining
        remaining[~problem_remaining] = False
        remaining = remaining.reshape(-1)
        improvement = torch.zeros(NM, device=q.device, dtype=q.dtype)
        for i in range(self.max_iterations):
            if not remaining.any():
                break
            # try stepping with this learning rate
            q_new = q + lr.unsqueeze(1) * dq
            # evaluate the error using tensor API when available
            if fk_fn is not None and eef_frame_idx is not None:
                all_tf = fk_fn(q_new)
                m = all_tf[eef_frame_idx].view(-1, M, 4, 4)
            else:
                m = chain.forward_kinematics(q_new).get_matrix()
                m = m.view(-1, M, 4, 4)
            dx, pos_diff, rot_diff = delta_pose(m, target_pos, target_wxyz)
            err_new = dx.squeeze().norm(dim=-1)
            # check if it's better
            improvement = err - err_new
            improved = improvement > self.sufficient_decrease
            # if it's better, we're done for those
            # if it's not better, reduce the learning rate
            lr[~improved] *= self.decrease_factor
            remaining = remaining & ~improved

        improvement = improvement.reshape(-1, M)
        improvement = improvement.mean(dim=1)
        return lr, improvement


class InverseKinematics:
    """Jacobian follower based inverse kinematics solver"""

    def __init__(self, serial_chain: SerialChain,
                 pos_tolerance: float = 1e-3, rot_tolerance: float = 1e-2,
                 retry_configs: Optional[torch.Tensor] = None, num_retries: Optional[int] = None,
                 joint_limits: Optional[torch.Tensor] = None,
                 config_sampling_method: Union[str, Callable[[int], torch.Tensor]] = "uniform",
                 max_iterations: int = 50,
                 lr: float = 1.0, line_search: Optional[LineSearch] = None,
                 regularlization: float = 1e-9, lm_damping: float = 0.1,
                 position_weight: float = 1.0, orientation_weight: float = 1.0,
                 debug=False,
                 early_stopping_any_converged=False,
                 early_stopping_no_improvement="any", early_stopping_no_improvement_patience=2,
                 enforce_joint_limits: bool = True,
                 num_limit_refinement_iterations: int = 10,
                 clamp_to_limits: bool = False
                 ):
        """
        :param serial_chain:
        :param pos_tolerance: position tolerance in meters
        :param rot_tolerance: rotation tolerance in radians
        :param retry_configs: (M, DOF) tensor of initial configs to try for each problem; leave as None to sample
        :param num_retries: number, M, of random initial configs to try for that problem; implemented with batching
        :param joint_limits: (DOF, 2) tensor of joint limits (min, max) for each joint in radians
        :param config_sampling_method: either "uniform" or "gaussian" or a function that takes in the number of configs
        :param max_iterations: maximum number of iterations to run
        :param lr: learning rate
        :param line_search: LineSearch object to use for line search
        :param regularlization: regularization term to add to the Jacobian
        :param debug: whether to print debug information
        :param early_stopping_any_converged: whether to stop when any of the retries for a problem converged
        :param early_stopping_no_improvement: {None, "all", "any", ratio} whether to stop when no improvement is made
        (consecutive iterations no improvement in minimum error - number of consecutive iterations is the patience).
        None means no early stopping from this, "all" means stop when all retries for that problem makes no improvement,
        "any" means stop when any of the retries for that problem makes no improvement, and ratio means stop when
        the ratio (between 0 and 1) of the number of retries that is making improvement falls below the ratio.
        So "all" is equivalent to ratio=0.999, and "any" is equivalent to ratio=0.001
        :param early_stopping_no_improvement_patience: number of consecutive iterations with no improvement before
        considering it no improvement
        :param position_weight: weight for position error in DLS objective. Higher values prioritize
        position accuracy. Default 1.0.
        :param orientation_weight: weight for orientation error in DLS objective. Higher values prioritize
        orientation accuracy. Default 1.0.
        :param lm_damping: Levenberg-Marquardt damping factor. Adds error-proportional regularization
        mu = lm_damping * ||error||^2 to the DLS solve, so large errors get more damping (stable) and
        small errors get less (fast convergence). Default 0.1.
        :param enforce_joint_limits: whether to enforce joint limits on the solution. Uses the chain's joint limits
        (chain.low/chain.high). After solving, revolute joints are wrapped by multiples of 2*pi, then any remaining
        violations are resolved via clamped refinement iterations. Set to False to disable.
        :param num_limit_refinement_iterations: number of clamped IK refinement iterations for enforcing joint limits
        :param clamp_to_limits: if True, clamp joint values to chain limits after each IK step (projected gradient).
        Uses chain.low/chain.high. Requires the chain to have finite joint limits. Default False.
        """
        self.chain = serial_chain
        self.dtype = serial_chain.dtype
        self.device = serial_chain.device
        joint_names = self.chain.get_joint_parameter_names(exclude_fixed=True)
        self.dof = len(joint_names)
        self.debug = debug
        self.enforce_joint_limits = enforce_joint_limits
        self.num_limit_refinement_iterations = num_limit_refinement_iterations
        self.clamp_to_limits = clamp_to_limits and torch.isfinite(serial_chain.low).any()
        # precompute which joints are revolute (for 2*pi wrapping)
        joints = self.chain.get_joints(exclude_fixed=True)
        self._revolute_mask = torch.tensor([j.joint_type == 'revolute' for j in joints],
                                           dtype=torch.bool, device=self.device)
        self.early_stopping_any_converged = early_stopping_any_converged
        self.early_stopping_no_improvement = early_stopping_no_improvement
        self.early_stopping_no_improvement_patience = early_stopping_no_improvement_patience

        self.max_iterations = max_iterations
        self.lr = lr
        self.regularlization = regularlization
        self.lm_damping = lm_damping
        self.line_search = line_search
        # Per-coordinate weighting: scales rows of Jacobian and error
        self._task_weight = torch.tensor(
            [position_weight] * 3 + [orientation_weight] * 3,
            device=self.device, dtype=self.dtype
        )  # (6,)

        self.err = None
        self.err_all = None
        self.err_min = None
        self.no_improve_counter = None

        self.pos_tolerance = pos_tolerance
        self.rot_tolerance = rot_tolerance
        self.initial_config = retry_configs
        if retry_configs is None and num_retries is None:
            raise ValueError("either initial_configs or num_retries must be specified")

        # sample initial configs instead
        self.config_sampling_method = config_sampling_method
        self.joint_limits = joint_limits
        if retry_configs is None:
            self.initial_config = self.sample_configs(num_retries)
        else:
            if retry_configs.shape[1] != self.dof:
                raise ValueError("initial_configs must have shape (N, %d)" % self.dof)
        # could give a batch of initial configs
        self.num_retries = self.initial_config.shape[-2]

    def clear(self):
        self.err = None
        self.err_all = None
        self.err_min = None
        self.no_improve_counter = None

    def sample_configs(self, num_configs: int) -> torch.Tensor:
        if self.config_sampling_method == "uniform":
            # bound by joint_limits
            if self.joint_limits is None:
                raise ValueError("joint_limits must be specified if config_sampling_method is uniform")
            return torch.rand(num_configs, self.dof, device=self.device) * (
                    self.joint_limits[:, 1] - self.joint_limits[:, 0]) + self.joint_limits[:, 0]
        elif self.config_sampling_method == "gaussian":
            return torch.randn(num_configs, self.dof, device=self.device)
        elif callable(self.config_sampling_method):
            return self.config_sampling_method(num_configs)
        else:
            raise ValueError("invalid config_sampling_method %s" % self.config_sampling_method)

    def solve(self, target_poses: Transform3d) -> IKSolution:
        """
        Solve IK for the given target poses in robot frame
        :param target_poses: (N, 4, 4) tensor, goal pose in robot frame
        :return: IKSolution solutions
        """
        raise NotImplementedError()


def delta_pose(m: torch.tensor, target_pos, target_wxyz, out: torch.Tensor = None):
    """
    Determine the error in position and rotation between the given poses and the target poses

    :param m: (N x M x 4 x 4) tensor of homogenous transforms
    :param target_pos:
    :param target_wxyz: target orientation represented in unit quaternion
    :param out: optional pre-allocated output buffer (N*M, 6, 1) to reduce memory allocation
    :return: (N*M, 6, 1) tensor of delta pose (dx, dy, dz, droll, dpitch, dyaw)
    """
    pos_diff = target_pos.unsqueeze(1) - m[:, :, :3, 3]
    pos_diff = pos_diff.view(-1, 3, 1)
    cur_wxyz = rotation_conversions.matrix_to_quaternion(m[:, :, :3, :3])

    # quaternion that rotates from the current orientation to the desired orientation
    # inverse for unit quaternion is the conjugate
    diff_wxyz = rotation_conversions.quaternion_multiply(target_wxyz.unsqueeze(1),
                                                         rotation_conversions.quaternion_invert(cur_wxyz))
    # angular velocity vector needed to correct the orientation
    # if time is considered, should divide by \delta t, but doing it iteratively we can choose delta t to be 1
    diff_axis_angle = rotation_conversions.quaternion_to_axis_angle(diff_wxyz)

    rot_diff = diff_axis_angle.view(-1, 3, 1)

    # Use pre-allocated buffer if provided
    if out is not None:
        out[:, :3] = pos_diff
        out[:, 3:] = rot_diff
        dx = out
    else:
        dx = torch.cat((pos_diff, rot_diff), dim=1)
    return dx, pos_diff, rot_diff


class PseudoInverseIK(InverseKinematics):
    def __init__(self, *args, use_compile: bool = False, **kwargs):
        """
        Initialize PseudoInverseIK solver.

        Args:
            *args: Arguments passed to InverseKinematics.
            use_compile: If True and PyTorch 2.0+ is available, use torch.compile
                for JIT compilation of FK, Jacobian, and IK step kernels. This can
                provide performance improvements after a warmup period. Default: False.
            **kwargs: Keyword arguments passed to InverseKinematics.
        """
        super().__init__(*args, **kwargs)
        # Pre-compute regularization matrix once
        self._reg_matrix = self.regularlization * torch.eye(6, device=self.device, dtype=self.dtype)

        # Set up compute kernels (potentially compiled)
        self._use_compile = use_compile and _TORCH_COMPILE_AVAILABLE
        if self._use_compile:
            self._ik_step_fn = torch.compile(_ik_step_kernel)
            self._jacobian_fn = torch.compile(self.chain.jacobian_tensor)
            self._fk_fn = torch.compile(self.chain.forward_kinematics_tensor)
        else:
            self._ik_step_fn = _ik_step_kernel
            self._jacobian_fn = self.chain.jacobian_tensor
            self._fk_fn = self.chain.forward_kinematics_tensor
        self._eef_frame_idx = self.chain._serial_eef_frame_idx

    def solve(self, target_poses: Transform3d) -> IKSolution:
        self.clear()

        target = target_poses.get_matrix()

        M = target.shape[0]

        target_pos = target[:, :3, 3]
        # jacobian gives angular rotation about x,y,z axis of the base frame
        # convert target rot to desired rotation about x,y,z
        target_wxyz = rotation_conversions.matrix_to_quaternion(target[:, :3, :3])

        sol = IKSolution(self.dof, M, self.num_retries, self.pos_tolerance, self.rot_tolerance,
                         device=self.device, dtype=self.dtype)

        q = self.initial_config
        if q.numel() == M * self.dof * self.num_retries:
            q = q.reshape(-1, self.dof)
        elif q.numel() == self.dof * self.num_retries:
            # repeat and manually flatten it
            q = self.initial_config.repeat(M, 1)
        elif q.numel() == self.dof:
            q = q.unsqueeze(0).repeat(M * self.num_retries, 1)
        else:
            raise ValueError(
                f"initial_config must have shape ({M}, {self.num_retries}, {self.dof}) or ({self.num_retries}, {self.dof})")
        # for logging, let's keep track of the joint angles at each iteration
        if self.debug:
            pos_errors = []
            rot_errors = []

        for i in range(self.max_iterations):
            with torch.no_grad():
                # early termination if we're out of problems to solve
                if not sol.remaining.any():
                    break
                sol.iterations += 1
                # compute Jacobian, end-effector pose, and pose error in one FK pass
                J, m = self._jacobian_fn(q, True)
                dq, dx = self._ik_step_fn(m, target_pos, target_wxyz, J,
                                          self._reg_matrix, self.num_retries, self.lm_damping, self._task_weight)
                dq = dq.squeeze(2)

                # convergence check using error at current q (before stepping)
                self.err_all = dx.squeeze(2)
                self.err = self.err_all.reshape(-1, self.num_retries, 6).norm(dim=-1)
                sol.update(q, self.err_all, use_keep_mask=self.early_stopping_any_converged)

                if self.early_stopping_no_improvement is not None:
                    if self.no_improve_counter is None:
                        self.no_improve_counter = torch.zeros_like(self.err)
                        self.err_min = self.err.clone()
                    else:
                        improved = self.err < self.err_min
                        self.err_min = torch.where(improved, self.err, self.err_min)

                        self.no_improve_counter = torch.where(improved,
                                                              torch.zeros_like(self.no_improve_counter),
                                                              self.no_improve_counter + 1)

                        # those that haven't improved
                        could_improve = self.no_improve_counter <= self.early_stopping_no_improvement_patience
                        # consider problems, and only throw out those whose all retries cannot be improved
                        could_improve = could_improve.reshape(-1, self.num_retries)
                        if self.early_stopping_no_improvement == "all":
                            could_improve = could_improve.all(dim=1)
                        elif self.early_stopping_no_improvement == "any":
                            could_improve = could_improve.any(dim=1)
                        elif isinstance(self.early_stopping_no_improvement, float):
                            ratio_improved = could_improve.sum(dim=1) / self.num_retries
                            could_improve = ratio_improved > self.early_stopping_no_improvement
                        sol.update_remaining_with_keep_mask(could_improve)

                if self.debug:
                    pos_err = dx[:, :3, 0].reshape(-1, 3).norm(dim=1)
                    rot_err = dx[:, 3:, 0].reshape(-1, 3).norm(dim=1)
                    pos_errors.append(pos_err)
                    rot_errors.append(rot_err)

            with torch.no_grad():
                if self.line_search is not None:
                    lr, _ = self.line_search.do_line_search(self.chain, q, dq, target_pos, target_wxyz,
                                                            dx, problem_remaining=sol.remaining,
                                                            fk_fn=self._fk_fn,
                                                            eef_frame_idx=self._eef_frame_idx)
                    lr = lr.unsqueeze(1)
                else:
                    lr = self.lr
                q = q.add(dq, alpha=lr) if isinstance(lr, float) else q.add(lr * dq)
                if self.clamp_to_limits:
                    q = torch.clamp(q, self.chain.low, self.chain.high)

        # Final convergence check for the last stepped q
        with torch.no_grad():
            J_final, m_final = self._jacobian_fn(q, True)
            _, dx_final = self._ik_step_fn(m_final, target_pos, target_wxyz, J_final,
                                           self._reg_matrix, self.num_retries, self.lm_damping, self._task_weight)
            self.err_all = dx_final.squeeze(2)
            self.err = self.err_all.reshape(-1, self.num_retries, 6).norm(dim=-1)
            sol.update(q, self.err_all, use_keep_mask=self.early_stopping_any_converged)

        if self.debug:
            # errors
            fig, ax = plt.subplots(ncols=2, figsize=(10, 5))
            pos_e = torch.stack(pos_errors, dim=0).cpu()
            rot_e = torch.stack(rot_errors, dim=0).cpu()
            ax[0].set_ylim(0, 1.)
            # ignore nan
            ignore = torch.isnan(rot_e)
            axis_max = rot_e[~ignore].max().item()
            ax[1].set_ylim(0, axis_max * 1.1)
            ax[0].set_xlim(0, self.max_iterations - 1)
            ax[1].set_xlim(0, self.max_iterations - 1)
            # draw at most 50 lines
            draw_max = min(50, pos_e.shape[1])
            for b in range(draw_max):
                c = (b + 1) / draw_max
                ax[0].plot(pos_e[:, b], c=cm.GnBu(c))
                ax[1].plot(rot_e[:, b], c=cm.GnBu(c))
            # label these axis
            ax[0].set_ylabel("position error")
            ax[1].set_xlabel("iteration")
            ax[1].set_ylabel("rotation error")
            plt.show()

        if self.enforce_joint_limits and self._has_finite_limits():
            self._enforce_limits(sol, target_pos, target_wxyz)

        return sol

    def _has_finite_limits(self):
        """Check if the chain has any finite joint limits."""
        return torch.isfinite(self.chain.low).any() or torch.isfinite(self.chain.high).any()

    def _wrap_revolute_joints(self, q):
        """Wrap revolute joint values by multiples of 2*pi to bring them closer to the valid range.

        For revolute joints, q and q + 2*n*pi produce identical FK, so we can freely shift
        by multiples of 2*pi without affecting the solution. Prismatic joints are left unchanged.
        """
        low = self.chain.low
        high = self.chain.high
        center = (low + high) / 2
        # compute the number of 2*pi shifts needed to bring q closest to the center of the limits
        n = torch.round((center - q) / (2 * math.pi))
        # only apply to revolute joints
        q_wrapped = q.clone()
        q_wrapped[..., self._revolute_mask] = q[..., self._revolute_mask] + n[..., self._revolute_mask] * 2 * math.pi
        return q_wrapped

    def _enforce_limits(self, sol, target_pos, target_wxyz):
        """Enforce joint limits on IK solutions via wrapping and clamped refinement.

        1. Wrap revolute joints by 2*pi to bring them into the valid range (preserves FK exactly).
        2. Clamp any remaining violations and run refinement iterations to recover convergence.
        3. Update the solution with the refined joint values.
        """
        q = sol.solutions.reshape(-1, self.dof)

        # Step 1: wrap revolute joints by 2*pi
        q = self._wrap_revolute_joints(q)

        # Step 2: clamp to limits
        q = torch.clamp(q, self.chain.low, self.chain.high)

        # Step 3: clamped refinement iterations to recover from any error introduced by clamping
        for _ in range(self.num_limit_refinement_iterations):
            J, m_flat = self._jacobian_fn(q, True)
            dq, _ = self._ik_step_fn(m_flat, target_pos, target_wxyz, J,
                                     self._reg_matrix, self.num_retries, self.lm_damping, self._task_weight)
            q = q + self.lr * dq.squeeze(2)
            q = torch.clamp(q, self.chain.low, self.chain.high)

        # Reset convergence so all retries are re-evaluated with the limit-enforced solutions
        sol.converged[:] = False
        sol.converged_any[:] = False
        sol.converged_pos[:] = False
        sol.converged_rot[:] = False

        # Recompute error and update solution
        J_new, m_new = self._jacobian_fn(q, True)
        _, dx_new = self._ik_step_fn(m_new, target_pos, target_wxyz, J_new,
                                     self._reg_matrix, self.num_retries, self.lm_damping, self._task_weight)
        sol.update(q, dx_new.squeeze(2), use_keep_mask=False)


class PseudoInverseIKWithSVD(PseudoInverseIK):
    """SVD-based damped least squares IK solver.

    About 2x slower per iteration than the default Cholesky-based solver (DLS),
    but converges slightly better near singularities (e.g. 99-100% vs 97-98%
    on near-singular targets). Exposes singular values which enables selective
    damping if subclassed further.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Override the step kernel with the SVD variant
        if self._use_compile:
            self._ik_step_fn = torch.compile(_ik_step_kernel_svd)
        else:
            self._ik_step_fn = _ik_step_kernel_svd



def calc_jacobian(serial_chain, th, tool=None, ret_eef_pose=False):
    """
    Return robot Jacobian J in base frame (N,6,DOF) where dot{x} = J dot{q}.
    Delegates to serial_chain.jacobian().
    """
    return serial_chain.jacobian(th, locations=tool, ret_eef_pose=ret_eef_pose)
